import React, { createContext, useState, useRef, useContext, useMemo, useEffect } from 'react';

function removeKeysFromObject(obj, keys) {
    return Object.keys(obj)
        .filter((key) => !keys.includes(key))
        .reduce((acc, key) => {
        acc[key] = obj[key];
        return acc;
    }, {});
}

const baseSerializeQueryParam = (value) => {
    if (isInvalidValue(value)) {
        return undefined;
    }
    if (Array.isArray(value)) {
        return value
            .filter((v) => !isInvalidValue(v))
            .map((v) => (typeof v === 'object' ? JSON.stringify(v) : v.toString()));
    }
    if (value instanceof Date) {
        return value.toISOString();
    }
    if (typeof value === 'object') {
        return JSON.stringify(value, function (_key, value) {
            if (typeof value === 'number' && isNaN(value)) {
                return 'NaN';
            }
            if (value === Infinity) {
                return 'Infinity';
            }
            if (value === -Infinity) {
                return '-Infinity';
            }
            if (value !== null && isInvalidValue(value)) {
                return undefined;
            }
            if (typeof value === 'object' &&
                !isEmptyObject(value) &&
                JSON.stringify(value) === '{}') {
                return undefined;
            }
            return value;
        });
    }
    return value === null || value === void 0 ? void 0 : value.toString();
};
function isInvalidValue(value) {
    return (value === null ||
        value === undefined ||
        (typeof value === 'number' && isNaN(value)) ||
        typeof value === 'symbol' ||
        typeof value === 'function' ||
        typeof value === 'bigint');
}
function isEmptyObject(value) {
    if (typeof value !== 'object' || value === null || Array.isArray(value)) {
        return false;
    }
    if (Object.prototype.toString.call(value) !== '[object Object]') {
        return false;
    }
    return Object.keys(value).length === 0;
}

function serializeQueryParamStates(states, customSerializeQueryParam) {
    const queryParams = {};
    for (const key of Object.keys(states)) {
        const state = states[key];
        const value = state.value;
        let serializedValue;
        if (state.serialize) {
            serializedValue = state.serialize(value);
        }
        else {
            serializedValue = customSerializeQueryParam
                ? customSerializeQueryParam(value)
                : baseSerializeQueryParam(value);
        }
        if (serializedValue !== undefined) {
            queryParams[key] = serializedValue;
        }
    }
    return queryParams;
}

function areUrlQueryValuesEqual(a, b) {
    if (Array.isArray(a) && Array.isArray(b)) {
        return a.length === b.length && a.every((value, index) => value === b[index]);
    }
    else if (typeof a === 'string' && typeof b === 'string') {
        return a === b;
    }
    return false;
}

function areUrlQueriesEqual(a, b) {
    const aKeys = Object.keys(a);
    const bKeys = Object.keys(b);
    if (aKeys.length !== bKeys.length || !aKeys.every((key) => bKeys.includes(key))) {
        return false;
    }
    return aKeys.every((key) => {
        const aValue = a[key];
        const bValue = b[key];
        return areUrlQueryValuesEqual(aValue, bValue);
    });
}

const NextQueryParamsContext = createContext(null);
/**
 * The `NextQueryParamsProvider` sets up a shared router adapter for `useNextQueryParams` hooks used within its children.
 *
 * @param props - the children and the adapter {@link NextQueryParamsAdapter}
 *
 * @example
 * ```tsx
 * function App({ Component, pageProps }: AppProps) {
 *  const router = useRouter();
 *  const adapter = new NextQueryParamsAdapter(router);
 *  return (
 *    <NextQueryParamsProvider adapter={adapter}>
 *        <Component {...pageProps} />
 *    </NextQueryParamsProvider>
 *  );
 * }
 * ```
 */
const NextQueryParamsProvider = ({ children, adapter }) => {
    return (React.createElement(NextQueryParamsContext.Provider, { value: {
            adapter
        } }, children));
};

function getChangedUrlQueryKeys(a, b) {
    const aKeys = Object.keys(a);
    const bKeys = Object.keys(b);
    const allKeys = Array.from(new Set([...aKeys, ...bKeys]));
    return allKeys.filter((key) => {
        const aValue = a[key];
        const bValue = b[key];
        return (aValue === undefined || bValue === undefined || !areUrlQueryValuesEqual(aValue, bValue));
    });
}

/**
 * The `useNextQueryParams` hook registers query param states and updates the URL query params when the states change.
 *
 * @param params - The query param states to register. {@link NextQueryParams}
 * @param adapter - The router adapter to use for interacting with the URL query params {@link NextQueryParamsAdapter}. If not provided, the adapter from the nearest provider {@link NextQueryParamsProvider} will be used.
 *
 * @example
 * ```tsx
 * function MyComponent() {
 *   const [str, setStr] = useState('world');
 *
 *   useNextQueryParams({
 *    hello: createStrQueryParam({
 *      value: str,
 *      onChange: setStr,
 *    })
 *   });
 *
 *   return <div>Hello {str}</div>;
 * }
 * ```
 */
function useNextQueryParams(params, adapter) {
    const [init, setInit] = useState(true);
    const [isStable, setStable] = useState(false);
    const [prevUrlQuery, setPrevUrlQuery] = useState({});
    const previouslyHasUrlQueryChanged = useRef(false);
    const keys = Object.keys(params);
    const nextQueryParamsContextValues = useContext(NextQueryParamsContext);
    const contextAdapter = nextQueryParamsContextValues === null || nextQueryParamsContextValues === void 0 ? void 0 : nextQueryParamsContextValues.adapter;
    const isRouterReady = (adapter === null || adapter === void 0 ? void 0 : adapter.isRouterReady) !== undefined
        ? adapter.isRouterReady
        : !!(contextAdapter === null || contextAdapter === void 0 ? void 0 : contextAdapter.isRouterReady);
    const urlQuery = (adapter === null || adapter === void 0 ? void 0 : adapter.urlQuery) || (contextAdapter === null || contextAdapter === void 0 ? void 0 : contextAdapter.urlQuery);
    const onStateChange = (adapter === null || adapter === void 0 ? void 0 : adapter.onChange) || (contextAdapter === null || contextAdapter === void 0 ? void 0 : contextAdapter.onChange);
    const mode = (adapter === null || adapter === void 0 ? void 0 : adapter.mode) || (contextAdapter === null || contextAdapter === void 0 ? void 0 : contextAdapter.mode) || 'reset';
    const customSerializeQueryParam = (adapter === null || adapter === void 0 ? void 0 : adapter.customSerializeQueryParam) || (contextAdapter === null || contextAdapter === void 0 ? void 0 : contextAdapter.customSerializeQueryParam);
    if (!onStateChange || !urlQuery) {
        throw new Error('useNextQueryParams is used outside a NextQueryParamsProvider, but no `onChange` function or `urlQuery` was passed to the hook.');
    }
    const uncontrolledKeys = Object.keys(urlQuery).filter((key) => !keys.includes(key));
    const queryParamStates = useMemo(() => {
        return Object.entries(params).reduce((acc, [key, value]) => {
            acc[key] = {
                value: value.value !== undefined && value.value !== null ? value.value : undefined,
                serialize: value.serialize
            };
            return acc;
        }, {});
    }, [params]);
    const hasUrlQueryChanged = useMemo(() => {
        return !areUrlQueriesEqual(urlQuery, prevUrlQuery);
    }, [urlQuery, prevUrlQuery]);
    const haveQueryParamStatesChanged = useMemo(() => {
        return !areUrlQueriesEqual(removeKeysFromObject(urlQuery, uncontrolledKeys), serializeQueryParamStates(queryParamStates, customSerializeQueryParam));
    }, [queryParamStates, urlQuery]);
    const onInit = () => {
        for (const key of keys) {
            const value = urlQuery[key];
            value && params[key].onChange(value);
        }
        previouslyHasUrlQueryChanged.current = true;
        setPrevUrlQuery(urlQuery);
        setInit(false);
    };
    const onChange = () => {
        var _a;
        if (hasUrlQueryChanged) {
            const changedKeys = getChangedUrlQueryKeys(urlQuery, prevUrlQuery);
            for (const key of changedKeys) {
                if (keys.includes(key)) {
                    const value = urlQuery[key];
                    const param = params[key];
                    if (value === undefined) {
                        if (mode === 'reset') {
                            param.onReset();
                        }
                    }
                    else {
                        const serializedStateValue = ((_a = param === null || param === void 0 ? void 0 : param.serialize) === null || _a === void 0 ? void 0 : _a.call(param, param === null || param === void 0 ? void 0 : param.value)) ||
                            (customSerializeQueryParam === null || customSerializeQueryParam === void 0 ? void 0 : customSerializeQueryParam(param === null || param === void 0 ? void 0 : param.value)) ||
                            baseSerializeQueryParam(param === null || param === void 0 ? void 0 : param.value);
                        if (!serializedStateValue ||
                            !areUrlQueryValuesEqual(serializedStateValue, value)) {
                            param.onChange(value);
                            previouslyHasUrlQueryChanged.current = true;
                        }
                    }
                }
            }
            setPrevUrlQuery(urlQuery);
        }
        else if (haveQueryParamStatesChanged) {
            onStateChange({
                ...removeKeysFromObject(urlQuery, keys),
                ...serializeQueryParamStates(queryParamStates, customSerializeQueryParam)
            }, previouslyHasUrlQueryChanged.current);
            previouslyHasUrlQueryChanged.current = false;
        }
        else {
            previouslyHasUrlQueryChanged.current = false;
        }
    };
    useEffect(() => {
        if (isRouterReady) {
            if (init) {
                onInit();
            }
            else {
                !haveQueryParamStatesChanged && setStable(true);
                onChange();
            }
        }
    }, [init, isRouterReady, hasUrlQueryChanged, haveQueryParamStatesChanged]);
    return { isStable };
}

/**
 * Creates a query param builder function.
 *
 * @param implementation - The implementation function of the query param builder function
 * @remarks
 * The implementation function accepts the props for a query param builder function {@link CreateQueryParamProps}
 * and returns a query param object that can be used with the `useQueryParams` hook. {@link NextQueryParam}
 *
 */
function createQueryParamFunctionFactory(implementation) {
    return function (props) {
        return implementation(props);
    };
}

/**
 * Builds a query param for a string value.
 *
 * @param props - The props for the query param function {@link CreateQueryParamProps}
 *
 * @returns A query param to be used in the `useQueryParams` hook {@link NextQueryParam}
 *
 * @example
 * ```tsx
 * function MyComponent() {
 *   const [str, setStr] = useState('foo');
 *   useNextQueryParams({
 *     str: createStrQueryParam({
 *       value: str,
 *       onChange: setStr,
 *     }),
 *   });
 *   return <div>Hello {str}</div>;
 *   // ?str=foo
 * }
 * ```
 */
const createStrQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize || ((v) => (v === null || v === undefined ? undefined : v)),
    onChange: (v) => {
        if (props.deserialize) {
            props.onChange(props.deserialize(v));
            return;
        }
        if (Array.isArray(v)) {
            props.onChange(v[0]);
        }
        else {
            props.onChange(v);
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue !== undefined
            ? props.defaultValue
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : '';
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));

/**
 * Builds a query param for a number value.
 *
 * @param props - The props for the query param function {@link CreateQueryParamProps}
 *
 * @returns A query param to be used in the `useQueryParams` hook {@link NextQueryParam}
 *
 * @example
 * ```tsx
 * function MyComponent() {
 *  const [num, setNum] = useState(0);
 *  useNextQueryParams({
 *    num: createNumQueryParam({
 *      value: num,
 *      onChange: setNum,
 *    }),
 *  });
 *  return <div>Hello {num}</div>;
 *  // ?num=0
 * }
 * ```
 */
const createNumQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize,
    onChange: (v) => {
        if (props.deserialize) {
            props.onChange(props.deserialize(v));
            return;
        }
        if (Array.isArray(v)) {
            if (!isNaN(Number(v[0]))) {
                props.onChange(Number(v[0]));
            }
        }
        else {
            if (!isNaN(Number(v))) {
                props.onChange(Number(v));
            }
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue !== undefined
            ? props.defaultValue
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : 0;
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));

/**
 * Builds a query param for a boolean value.
 *
 * @param props - The props for the query param function {@link CreateQueryParamProps}
 *
 * @returns A query param to be used in the `useQueryParams` hook {@link NextQueryParam}
 *
 * @example
 * ```tsx
 * function MyComponent() {
 *  const [bool, setBool] = useState(true);
 *  useNextQueryParams({
 *    bool: createBoolQueryParam({
 *      value: bool,
 *      onChange: setBool,
 *    }),
 *  });
 *  return <div>Hello {bool ? 'World' : 'Universe'}</div>;
 *  // ?bool=true
 * }
 * ```
 */
const createBoolQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize ||
        ((v) => {
            if (v === undefined || v === null) {
                return undefined;
            }
            return v.toString();
        }),
    onChange: (v) => {
        if (props.deserialize) {
            props.onChange(props.deserialize(v));
            return;
        }
        if (Array.isArray(v)) {
            if (v[0] === 'true') {
                props.onChange(true);
            }
            else if (v[0] === 'false') {
                props.onChange(false);
            }
        }
        else {
            if (v === 'true') {
                props.onChange(true);
            }
            else if (v === 'false') {
                props.onChange(false);
            }
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue !== undefined
            ? props.defaultValue
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : false;
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));

/**
 * Builds a query param for a JSON record value.
 *
 * @param props - The props for the query param function {@link CreateQueryParamProps}
 *
 * @returns A query param to be used in the `useQueryParams` hook {@link NextQueryParam}
 *
 * @example
 * ```tsx
 * function MyComponent() {
 *   // The explicit typing of `record` is necessary to ensure that the value is a JSON record
 *   const [record, setRecord] = useState<Record<string, string>>({ foo: 'bar' });
 *   useNextQueryParams({
 *     record: createJSONRecordQueryParam({
 *       value: record,
 *       onChange: setRecord,
 *     }),
 *   });
 *   return <div>Hello {record.foo}</div>;
 *   // ?record=%7B%22foo%22%3A%22bar%22%7D
 * }
 */
const createJSONRecordQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize ||
        ((v) => {
            if (v === undefined || v === null) {
                return undefined;
            }
            return JSON.stringify(v);
        }),
    onChange: (v) => {
        if (props.deserialize) {
            props.onChange(props.deserialize(v));
            return;
        }
        let parsed;
        try {
            if (Array.isArray(v)) {
                parsed = JSON.parse(v[0]);
            }
            else {
                parsed = JSON.parse(v);
            }
        }
        catch (e) {
            return;
        }
        if (isJSONRecord(parsed)) {
            props.onChange(parsed);
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue !== undefined
            ? props.defaultValue
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : {};
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));
function isJSONRecord(value) {
    return typeof value === 'object' && value !== null && !Array.isArray(value);
}

/**
 * Builds a query param for a date value.
 *
 * @param props - The props for the query param function {@link CreateQueryParamProps}
 *
 * @returns A query param to be used in the `useQueryParams` hook {@link NextQueryParam}
 *
 * @example
 * ```tsx
 * function MyComponent() {
 *  const [date, setDate] = useState(new Date('2021-01-01T00:00:00'));
 *  useNextQueryParams({
 *    date: createDateQueryParam({
 *      value: date,
 *      onChange: setDate,
 *      withTime: true,
 *    }),
 *    dateWithoutTime: createDateQueryParam({
 *      value: date,
 *      onChange: setDate,
 *    }),
 *  });
 *  return <div>Hello {date.toISOString()}</div>;
 *  // ?date=2021-01-01T00:00:00&dateWithoutTime=2021-01-01
 * }
 * ```
 */
const createDateQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize ||
        ((v) => {
            if (v === undefined || v === null) {
                return undefined;
            }
            if (props.withTime) {
                return v.toISOString().split('.')[0];
            }
            return v.toISOString().split('T')[0];
        }),
    onChange: (v) => {
        if (props.deserialize) {
            props.onChange(props.deserialize(v));
            return;
        }
        if (Array.isArray(v)) {
            const parsedDate = new Date(v[0]);
            if (!isNaN(parsedDate.getTime())) {
                props.onChange(parsedDate);
            }
        }
        else {
            const parsedDate = new Date(v);
            if (!isNaN(parsedDate.getTime())) {
                props.onChange(parsedDate);
            }
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue
            ? new Date(props.defaultValue)
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : new Date();
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));

/**
 * Builds a query param for a string array value.
 *
 * @param props - The props for the query param function {@link CreateQueryParamProps}
 *
 * @returns A query param to be used in the `useQueryParams` hook {@link NextQueryParam}
 *
 * @example
 * ```tsx
 * function MyComponent() {
 *   const [strArr, setStrArr] = useState(['foo', 'bar']);
 *   useNextQueryParams({
 *     strArr: createStrArrQueryParam({
 *       value: strArr,
 *       onChange: setStrArr,
 *     }),
 *   });
 *   return <div>Hello {strArr.join(', ')}</div>;
 *   // ?strArr=foo&strArr=bar
 * }
 */
const createStrArrQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize,
    onChange: (v) => {
        if (props.deserialize) {
            props.onChange(props.deserialize(v));
            return;
        }
        if (Array.isArray(v)) {
            props.onChange(v);
        }
        else {
            props.onChange([v]);
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue !== undefined
            ? props.defaultValue
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : [];
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));

/**
 * Builds a query param for a number array value.
 *
 * @param props - The props for the query param function {@link CreateQueryParamProps}
 *
 * @returns A query param to be used in the `useQueryParams` hook {@link NextQueryParam}
 *
 * @example
 * ```tsx
 * function MyComponent() {
 *   const [numArr, setNumArr] = useState([1, 2, 3]);
 *   useNextQueryParams({
 *     numArr: createNumArrQueryParam({
 *       value: numArr,
 *       onChange: setNumArr,
 *     }),
 *   });
 *   return <div>Hello {numArr.join(', ')}</div>;
 *   // ?numArr=1&numArr=2&numArr=3
 * }
 */
const createNumArrQueryParam = createQueryParamFunctionFactory((props) => ({
    value: props.value,
    serialize: props.serialize,
    onChange: (v) => {
        if (props.deserialize) {
            props.onChange(props.deserialize(v));
            return;
        }
        if (Array.isArray(v)) {
            props.onChange(v.filter((v) => !isNaN(Number(v))).map((v) => Number(v)));
        }
        else {
            if (!isNaN(Number(v))) {
                props.onChange([Number(v)]);
            }
        }
    },
    onReset: () => {
        const defaultValue = props.defaultValue !== undefined
            ? props.defaultValue
            : props.nullable
                ? null
                : props.optional
                    ? undefined
                    : [];
        /**
         * @note The usage of `any` in the following is necessary because TypeScript is unable to
         * infer the generic types N and O in `createQueryParamFunction` without exposing them, making it
         * impossible to maintain complete type safety for the `props.onChange` argument here.
         * To ensure correctness, be cautious and thoroughly test this code.
         */
        props.onChange(defaultValue);
    }
}));

/**
 * Creates a Next.js router adapter for the `useNextQueryParams` hook.
 *
 * @param nextRouter - The next router instance
 * @param options - The options for the adapter {@link NextRouterAdapterOptions}
 *
 * @example
 * ```tsx
 * function App({ Component, pageProps }: AppProps) {
 *   const router = useRouter();
 *   const adapter = createNextRouterAdapter(router);
 *   return (
 *     <NextQueryParamsProvider adapter={adapter}>
 *       <Component {...pageProps} />
 *     </NextQueryParamsProvider>
 *   );
 * }
 * ```
 */
function createNextRouterAdapter(nextRouter, options) {
    const { mode, replace, shallow, onChange } = options || {};
    const defaultOnChange = (query, isTriggeredByUrl) => {
        const updateMethod = replace || isTriggeredByUrl ? 'replace' : 'push';
        nextRouter[updateMethod]({
            pathname: nextRouter.pathname,
            query
        }, undefined, { shallow: !!shallow });
    };
    return {
        isRouterReady: nextRouter.isReady,
        urlQuery: nextRouter.query,
        mode,
        onChange: onChange || defaultOnChange
    };
}

export { NextQueryParamsProvider, createBoolQueryParam, createDateQueryParam, createJSONRecordQueryParam, createNextRouterAdapter, createNumArrQueryParam, createNumQueryParam, createQueryParamFunctionFactory, createStrArrQueryParam, createStrQueryParam, useNextQueryParams };
//# sourceMappingURL=index.esm.mjs.map
